/*
   piktd.c -- PIKT alert daemon

   $Header: /var/cvs/PIKT/pikt/src/piktd/piktd.c,v 1.67 2007/01/02 09:14:20 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2006 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org

   --------------------------------------------------------------------------

   Usage:  piktd [-GhTVz]

           -G              run in debug mode
           -h              show program help
           -T              run in test mode
           -V              show version info
           -z              run in 'make check' mode

 */

#include "../lib/common.h"
#include "piktd.h"

const char piktd_version[] = "PIKT piktd version " VERSION;

/* command-line options */
int ischeck = FALSE;
int isdebug = FALSE;
int isdoublecheck = FALSE;
int ishelp = FALSE;
int isshow_version = FALSE;
int istest = FALSE;
int ismakecheck = FALSE;

int bailout = FALSE;

const char nullstring[] = { '\0' };

int lineno = 0;

alt *alerts = (alt *) NULL;
alt *alert;

char *mailcmd, *lpcmd;

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

unsigned int fork_delay = 1;

alt *readconf(void);
void timingsync(time_t *);
void timingsleep(time_t);

int
main (int argc, char *argv[])
{
	static char fnnam[] = "main";
	pid_t pid;
	static char pidstr[8];
	struct tm *tm;
	alt *altp;
	unsigned long himincurr, himindrift;  /* 32 bits, for seconds 0-31 */
	unsigned long lomincurr, lomindrift;  /* 32 bits, for seconds 32-59 */
	unsigned long hrcurr,       hrdrift;  /* hour */
	unsigned long domcurr,     domdrift;  /* day of month */
	unsigned long moncurr,     mondrift;  /* month */
	unsigned long dowcurr,     dowdrift;  /* day of week */
	static time_t nextmin;
	static time_t driftmin;
	int launch;

	init(argc, argv, (void (*)(void))piktd_usage, piktd_version);
	if (bailout)
		exit(exstat);

	/* possibly check for cmd strings in the future, ala pikt.c */

	/* check for blocking lockfile */
	if (islock(piktdlock)) {
		logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTD, __FILE__, __LINE__, fnnam, FALSE, "%s found!  Another piktd may be running.  Try killing the running piktd (else rm the %s lockfile).", piktdlock, piktdlock);
	}
	/* check for running piktd */
	if (((pid = procid("piktd", PROCID)) == ERROR) || (pid != NIL)) {
		logexit(OUTPUT, LOG_ERR, ERRID_MULTIPLE_PIKTD, __FILE__, __LINE__, fnnam, FALSE, "Another piktd may be running.  Try killing the running piktd (and/or rm the %s lockfile, if any).", piktdlock);
	}

	/* go into daemon mode */
	daemon_init(TRUE);

	/* determine the daemon pid */
	pid = getpid();
	sprintf(pidstr, "%u", (unsigned)pid);

	/* apply the lock, to prevent other concurrent daemon sessions */
	if (mklock(piktdlock, pidstr, FALSE, TRUE) == ERROR)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CREATE_ERROR, __FILE__, __LINE__, fnnam, FALSE, "mklock() error for %s", piktdlock);

	if (! readconf())
		logexit(PIKTLOG, LOG_ERR, ERRID_PIKTDCONF_NO_ALERTS_ENABLED, __FILE__, __LINE__, fnnam, FALSE, "no alerts enabled, or error in %s", piktdconf);

	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "daemon (re)start");

	timingsync(&nextmin);	/* determine secs to next minute */

	while (1) {

		timingsleep(nextmin);	/* sleep to resync with next minute */

		/* get current time */
		tm = localtime(&nextmin);

		settimebits(tm, &himincurr, &lomincurr, &hrcurr, &domcurr, &moncurr, &dowcurr);

		/* for each alert, look for timing matches; if found
		   launch the alert
		 */
		foreach (altp, alerts) {

			launch = FALSE;

			if (altp->drift) {
				if (altp->timer == 0) {
					driftmin = nextmin + (altp->drift * 60);
					/* get forward drift time */
					tm = localtime(&driftmin);
					settimebits(tm, &himindrift, &lomindrift, &hrdrift, &domdrift, &mondrift, &dowdrift);
					if (timematch(altp, himindrift, lomindrift,
					                    hrdrift,    domdrift,
					                    mondrift,   dowdrift)
					   ) {
						/* 2*, because +/- drift */
						altp->timer = randomizetimer(2*(altp->drift));
						if (altp->timer == 0)
							launch = TRUE;
					}
				} else {
					altp->timer--;
					if (altp->timer == 0)
						launch = TRUE;
				}
			} else {	/* the normal case, no time drift */
				if (timematch(altp, himincurr, lomincurr,
				                    hrcurr,    domcurr,
				                    moncurr,   dowcurr)
				   ) {
					launch = TRUE;
				}
			}

			if (launch) {
				/* launch the alert, but only if no prior
				   lock file
				 */
				if (islock(altp->lockfile)) {
					logreturn(PIKTLOG, LOG_ERR, ERRID_MULTIPLE_PIKT, __FILE__, __LINE__, fnnam, FALSE, "%s found, another %s pikt run may be in progress, run aborted.", altp->lockfile, altp->name);
					continue;	/* skip it */
				}
				/* first we fork() a child; the parent waits
				   for the death of the child; the first
				   child forks()'s a grandchild, then
				   immediately dies; the parent checks the
				   exit status of the child then moves on;
				   meanwhile, it is the grandchild that does
				   the actual execl() of the alert command;
				   since its parent (the first child) has
				   died, init inherits the grandchild; in this
				   way, we avoid both zombies and occasional
				   mysterious piktd hangs
				 */
				switch (pid = fork()) {
					case -1:	/* error */
						logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE,  "fork() error");
						break;
					case 0:		/* child */
						switch (pid = fork()) {
							case -1:	/* error */
								logexit(PIKTLOG, LOG_ERR, ERRID_FORK_ERROR, __FILE__, __LINE__, fnnam, TRUE,  "fork() error");
								break;
							case 0:		/* grandchild */
								/* always log the pikt command */
								logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "%s", altp->piktcmd);
								/* but if in test mode,
								   don't exec the pikt command
								 */
								if (! istest) {
									execl(SHELL, "sh", "-c", altp->piktcmd, (char *) NULL);
								}
								exit(0);
								break;
							default:	/* child */
								exit(0);
								break;
						}
						break;
					default:	/* parent */
						break;
				}	/* switch (pid = fork()) */

			}	/* if (launch) */

		}	/* foreach (altp, alerts) */

		/* set ahead for next go-around */
		nextmin += 60;

	}	/* while (1) */

}

/* output program usage to screen
 */
void
piktd_usage (void)
{
	printf("\nUsage:  piktd [-GhTVz]\n\n");
	printf("\t-G\t\trun in debug mode\n");
	printf("\t-h\t\tshow program help\n");
	printf("\t-T\t\trun in test mode\n");
	printf("\t-V\t\tshow version info\n");
	printf("\t-z\t\trun in 'make check' mode\n");
	return;
}

/* read in the piktd.conf file; for each alert, determine the timing, and
   the indicated pikt command string
 */
alt *
readconf (void)
{
	static char fnnam[] = "readconf";
	FILE *conffile, *altfile;
	char line[MAXLINE+1];
	char minutes[MAXLINE+1], hours[MAXLINE+1], days_of_month[MAXLINE+1], months[MAXLINE+1], days_of_week[MAXLINE+1], drift[MAXLINE+1], cmd[MAXLINE+1];
	char *cp;
	char c;
	alt *naltp;

	altfile = (FILE *) NULL;
	if ((conffile = fopen(piktdconf, "r")) == NULL)
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdconf);
	line[MAXLINE-1] = '\0';
	while (fgets(line, MAXLINE+1, conffile) != NULL) {
		if (((c = line[MAXLINE-1]) != '\0') && (c != '\n'))
			logexit(PIKTLOG, LOG_ERR, ERRID_PIKTDCONF_LINE_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, line too long: %s", piktdconf, line);
		if ((cp = strchr(line, '\n')) != NULL)
			*cp = '\0';
		if (line[0] == '#')
			continue;
		if (strstr(line, PIKTNEVER))
			continue;
		if (   (sscanf(line, "%s %s %s %s %s %s %[^&]", minutes, hours, days_of_month, months, days_of_week, drift, cmd) != 7)
		    || (((cp = strstr(cmd, "+A ")) == NULL) && ((cp = strstr(cmd, "+S ")) == NULL))
		    || (cp && ! isalpha((int)*(cp+3)))
		   ) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_PIKTDCONF_LINE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "bad %s line: %s", piktdconf, line);
		} else {
			naltp = (alt *) mymalloc(sizeof(alt), fnnam);
			naltp->name = (char *) mymalloc((strlen(cp+3)+1) * sizeof(char), fnnam);
			strcpy(naltp->name, cp+3);
			/* to keep things simple, in the following, we
			   disregard the leap year complication, and we
			   assume a standard month of 30 days
			 */
			parsetiming(months,        &(naltp->monprob), &(naltp->mon),    1, 12,  1, 12, 12, 60*24*30, 60*24*365);
			parsetiming(days_of_week,  &(naltp->dowprob), &(naltp->dow),    0,  6,  0,  6,  7,    60*24,   60*24*7);
			parsetiming(days_of_month, &(naltp->domprob), &(naltp->dom),    1, 31,  1, 31, 31,    60*24,  60*24*30);
			parsetiming(hours,         &(naltp->hrprob),  &(naltp->hr),     0, 23,  0, 23, 24,       60,     60*24);
			parsetiming(minutes,       &(naltp->minprob), &(naltp->himin), 32, 59,  0, 59, 60,        1,        60);
			parsetiming(minutes,       &(naltp->minprob), &(naltp->lomin),  0, 31,  0, 59, 60,        1,        60);
			naltp->drift = atoi(drift);	/* we don't check for
							   error here, since
							   drift has been
							   checked before this
							 */
			if (2*(naltp->drift) >= mai) {
				logexit(PIKTLOG, LOG_ERR, ERRID_MIN_ALERT_INTERVAL_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "for alert %s, +/- drift %d equals or exceeds minimum alert interval %d", naltp->name, naltp->drift, mai);
			}
			naltp->timer = 0;
			naltp->lockfile = (char *) mymalloc((strlen(lokdir)+strlen(naltp->name)+11) * sizeof(char), fnnam);
			sprintf(naltp->lockfile, "%s/%s.alt.lock", lokdir, naltp->name);
			naltp->piktcmd = mystrdup(cmd, fnnam);
			naltp->next = (alt *) NULL;
			if (! alerts)
				alerts = naltp;
			else
				alert->next = naltp;
			alert = naltp;
			if (! altfile) {
				if ((altfile = fopen(piktdalt, "w")) == NULL)
					logexit(PIKTLOG, LOG_ERR, ERRID_FILE_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fopen() error for %s", piktdalt);
			}
			fprintf(altfile, "%s\n", naltp->name);
		}
		line[MAXLINE-1] = '\0';
	}
	if (altfile) {
		if (fclose(altfile))
			logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE,  "fclose() error for %s", piktdalt);
	}
	chmod(piktdalt, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));	/* 640 */
	if (fclose(conffile))
		logexit(PIKTLOG, LOG_ERR, ERRID_FILE_CLOSE_ERROR, __FILE__, __LINE__, fnnam, TRUE, "fclose() error for %s", piktdconf);
	return alerts;
}

/* determine the number of seconds until the next minute
 */
void
timingsync (time_t *nmp)
{
	register struct tm *tm;

	*nmp = time((time_t*)0);
	tm = localtime(nmp);
	*nmp += (60 - tm->tm_sec);
}

/* sleep so as to resync with the next minute
 */
void
timingsleep (time_t nm)
{
	register int waitsecs;

	waitsecs = (int) (nm - time((time_t*)0));
	while (waitsecs > 0)
		waitsecs = (int) sleep((unsigned int) waitsecs);
}
